//=======================================================================
// cftv.h : Closed-form Solution to Tensor Voting (CFTV) -- ND implementation.
//
// Copyright (c) 2009 Tai-Pang Wu (pang@cse.ust.hk) and HKUST x CUHK
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#ifndef _CFTV_H_
#define _CFTV_H_

#include "StdAfx.h"
#include "ANN/ANN.h"

// Global variables for cftv(). Simply for speed-up.
// I don't want to allocate and deallocate memory every time I call cftv.
// Please see the detail in cftv().
class CFTV {
public:
	static int          currDim;
	static double**     R;
	static double**     Rp;
	static double**     outr;
	static double*      r;
	static double**     I;
};

int        CFTV::currDim = -1;
double**   CFTV::R = NULL;
double**   CFTV::Rp = NULL;
double**   CFTV::outr = NULL;
double*    CFTV::r = NULL;
double**   CFTV::I = NULL;


// Closed-form Solution to Tensor Voting
// REMARK: This function is not limited to 3D. The dimension can be set by "dim". See following.
//
// dim             : Dimension of the point
// x1              : Location of point x1
// x2              : Location of point x2
// K2              : The tensor correspond to x2 (IT IS NOT LIMITED TO IDENTITY! It can be any tensor!)
// S[output]       : Estimated tensor at x1
// sigma           : Sigma of the decay function (i.e. \sigma_d of Eqn 2 (or Eqn 4 in reference [28]))
void cftv( int dim, double* x1, double* x2, double** K2, double** S, double sigma )
{

	// For convenience
	static int&       currDim = CFTV::currDim;
	static double**&  R       = CFTV::R;
	static double**&  Rp      = CFTV::Rp;
	static double**&  outr    = CFTV::outr;
	static double*&   r       = CFTV::r;
	static double**&  I       = CFTV::I;

	int i, j;

	// reallocate memory if dim is not the same
	if( currDim != dim )
	{
		safeDelete2DDouble( currDim, R );
		safeDelete2DDouble( currDim, Rp );
		safeDelete2DDouble( currDim, outr );
		SAFEDELARR( r );
		safeDelete2DDouble( currDim, I );

		R = allocate2DDouble( dim, dim );
		Rp = allocate2DDouble( dim, dim );
		outr = allocate2DDouble( dim, dim );
		r = new double[ dim ];
		I = allocate2DDouble( dim, dim );
		eye( dim, I );

		currDim = dim;
	}

	// ------------- CFTV start here -------------
	for( i = 0; i < dim; i++ )
		r[i] = x1[i] - x2[i];

	double length_r2 = norm2( dim, r );
	double length_r  = sqrt( length_r2 );

	if( !nearzero( length_r ) )
		for( i = 0; i < dim; i++ )
			r[i] = r[i] / length_r;

	// outer product of r
	outerProduct( dim, r, dim, r, outr ); 

	for( i = 0; i < dim; i++ )
		for( j = 0; j < dim; j++ )
			outr[i][j] = 2.0 * outr[i][j];

	// find R
	subMat( dim, dim, I, outr, R );

	for( i = 0; i < dim; i++ )
		for( j = 0; j < dim; j++ )
			outr[i][j] = outr[i][j] / 4.0;

	// find Rp
	subMat( dim, dim, I, outr, Rp );
	multiMat( dim, dim, dim, Rp, R, outr );
	copyMatrix( outr, Rp, dim, dim );

	//find S
	multiMat( dim, dim, dim, R, K2, outr );
	multiMat( dim, dim, dim, outr, Rp, S );
	
	double c = exp( - fabs( length_r2 ) / sigma );

	// scale the S by distance
	for( i = 0; i < dim; i++ )
		for( j = 0; j < dim; j++ )
			S[i][j] = c * S[i][j];

	// ------------- CFTV end here -------------
}


// Standard Tensor Voting (with CFTV)
// REMARK: This function is not limited to 3D. The dimension can be set by "dim". See following.
//
// K[input/output] : The set of tensors correspond to x (You can initialize them with any tensor (NOT LIMITED TO IDENTITY))
// x               : The set of input points
// numPoint        : Number of input point
// dim             : Dimension of the point
// sigma           : Sigma of the decay function (i.e. \sigma_d of Eqn 2 (or Eqn 4 in reference [28]))
// maxIter         : Maximum number of iteration
// epsBoundary     : The cutoff boundary for Eqn 2. Note that the tail of Gaussian (decay function) is long where 
//                   the values smaller than epsBoundary could be ignored.
int tensorVoting( double*** K, double** x, int numPoint, int dim, double sigma, 
					int maxIter, double epsBoundary = 0.0075)
{
	if( K == NULL ) return 0;
	if( x == NULL ) return 0;
	if( numPoint <= 0) return 0;
	if( dim <= 0 ) return 0;
	if( sigma < EPLSION ) return 0;

	int iter = 0;
	int i, j, k, m;

	double** S    = allocate2DDouble( dim, dim );
	double*** Kp = new double**[numPoint];
	for( i = 0; i < numPoint; i++ )
	{
		Kp[i] = allocate2DDouble( dim, dim );
		zeroMat( dim, dim, Kp[i] );
		for( j = 0; j < dim; j++ )
			Kp[i][j][j] = 0.00001;
	}

	// squared neighbor radius
	double radius2 = - sigma * log( epsBoundary );

	// create neighbor list
	printf( "    [CFTV Msg] Create neighbor list .... ");

	ANNkd_tree* annTree = new ANNkd_tree( (ANNpointArray) x, numPoint, dim );
	int** neighborList = new int*[numPoint];
	int* neighborNum = new int[numPoint];
	double* distArray = new double[numPoint];
	int maxNumNeighbor = -MYINFINITE;
	int minNumNeighbor = MYINFINITE;
	double avgNumNeighbor = 0;
	for( i = 0; i < numPoint; i++ )
	{
		int numNeighbor = numPoint;
		numNeighbor = annTree->annkFRSearch( (ANNpoint) x[i], (ANNdist) radius2, numNeighbor );
		int* tempIndexArray = new int[numNeighbor];
		numNeighbor = annTree->annkFRSearch( (ANNpoint) x[i], (ANNdist) radius2, numNeighbor, (ANNidxArray) tempIndexArray, (ANNdistArray) distArray );
		neighborList[i] = tempIndexArray;
		neighborNum[i] = numNeighbor;
		if( numNeighbor > maxNumNeighbor ) maxNumNeighbor = numNeighbor;
		if( numNeighbor < minNumNeighbor ) minNumNeighbor = numNeighbor;
		avgNumNeighbor += double(numNeighbor);
	}
	avgNumNeighbor = avgNumNeighbor / double(numPoint);
	SAFEDEL( annTree );
	SAFEDELARR( distArray );
	annClose();
	printf( "Done\n" );


	printf("    [CFTV Msg] Standard Tensor Voting .... " );
	for( iter = 0; iter < maxIter; iter++ )
	{
		// Perform CFTV for each point
		for (i = 0; i < numPoint; i++ )
		{
			// collect votes from neighbors of i
			for( j = 0; j < neighborNum[i]; j++ )
			{
				cftv( dim, x[i], x[neighborList[i][j]], K[neighborList[i][j]], S, sigma );
				addMat( dim, dim, Kp[i], S, Kp[i] );
			}
		}

		// copy result to the input/output 
		for( i = 0; i < numPoint; i++ )
			copyMatrix( Kp[i], K[i], dim, dim );
	}
	printf( "Done\n" );


	safeDelete2DDouble( dim, S );
	if( Kp != NULL )
	{
		for( i = 0; i < numPoint; i++ )
			safeDelete2DDouble( dim, Kp[i] );
		SAFEDELARR( Kp );
	}

	if( neighborList != NULL )
	{
		for( i = 0; i < numPoint; i++ )
			SAFEDELARR( neighborList[i] );
		SAFEDELARR( neighborList );
	}

	SAFEDELARR( neighborNum );

	return iter;
}


#endif

