//=======================================================================
// emtv.cpp : Definition of EMPlane class
//
// Copyright (c) 2009 Tai-Pang Wu (pang@cse.ust.hk) and HKUST x CUHK
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//=======================================================================

#include "StdAfx.h"
#include "emtv.h"
#include "svdcmp.h"
#include "eig.h"
#include "ANN/ANN.h"

#include <assert.h>

EMPlane::EMPlane( int dim, int maxIter, double epsBoundary )
{
	R = NULL;
	Rp = NULL;
	outr = NULL;
	r = NULL;
	I = NULL;

	m_fSigmaScale = 1;
	m_afSigma[0] = 0.01;
	m_afSigma[1] = 0.01;
	m_afSigma[2] = 0.01;

	m_iDim = dim;
	m_iMaxIter = maxIter;
	m_fEpsBoundary = epsBoundary; 
	m_fMEsp = 0.001;
}

EMPlane::EMPlane( int dim, double sigma_scale, double sigma0, double sigma1, double sigma2, int maxIter, double epsBoundary )
{
	R = NULL;
	Rp = NULL;
	outr = NULL;
	r = NULL;
	I = NULL;

	m_fSigmaScale = sigma_scale;
	m_afSigma[0] = sigma0;
	m_afSigma[1] = sigma1;
	m_afSigma[2] = sigma2;

	m_iDim = dim;
	m_iMaxIter = maxIter;
	m_fEpsBoundary = epsBoundary;
	m_fMEsp = 0.001;
}

EMPlane::~EMPlane()
{
	safeDelete2DDouble( m_iDim, R );
	safeDelete2DDouble( m_iDim, Rp );
	safeDelete2DDouble( m_iDim, outr );
	SAFEDELARR( r );
	safeDelete2DDouble( m_iDim, I );
}

int EMPlane::em_plan( double** x, int numPt, double*& v, double***& K, double*& w )
{

	assert( x != NULL );

	int i, j, k;

	SAFEDELARR( v );
	SAFEDELARR( w );
	if( K != NULL )
	{
		for( i =0; i< numPt; i++ )
			safeDelete2DDouble( m_iDim, K[i] );
		SAFEDELARR( K );
	}

	// allocate space for temporary storage
	double* nVec = new double[ m_iDim ];
	memset( nVec, 0, sizeof( double ) * m_iDim );

	// allocate space for output
	v = new double[ m_iDim ];
	memset( v, 0, sizeof( double ) * m_iDim );
	w = new double[ numPt ];
	memset( w, 0, sizeof( double ) * numPt );
	K = new double**[numPt];
	for( i = 0; i < numPt; i++ )
		K[i] = allocate2DDouble( m_iDim, m_iDim );
	
	double*** Kp = new double**[ numPt ];
	for( i = 0; i < numPt; i++ )
		Kp[i] = allocate2DDouble( m_iDim, m_iDim );


	// the paramter to be estimated in the algorithm
	double alpha = 0.9;
	double sigma0 = 100;
	double sigma1 = 100;
	double sigma2 = 100;

	// initialize K
	for( i = 0; i < numPt; i++ )
	{
		eye( m_iDim, K[i] );
		eye( m_iDim, Kp[i] );
	}

	// guess the initial v here
	initialize( numPt, v, x, K, 4 );

	printf( "initial guess of v = [ " );
	for( i = 0; i < m_iDim; i++ )
		printf( "%f ", v[i] );
	printf ("]\n" );

	// create ANN kd-tree
	ANNkd_tree annTree( (ANNpointArray) x, numPt, m_iDim );

	// create temporary storage
	int* indexArray = new int[ numPt ];
	double* distArray = new double[ numPt ];
	double** newK = allocate2DDouble( m_iDim, m_iDim );
	double** tempS = allocate2DDouble( m_iDim, m_iDim );
	double** tempvv = allocate2DDouble( m_iDim, m_iDim );
	double** M = allocate2DDouble( m_iDim, m_iDim );
	double** G = allocate2DDouble( m_iDim, m_iDim );
	double** tempK = allocate2DDouble( m_iDim, m_iDim );
	double** eigVec = allocate2DDouble( m_iDim, m_iDim );
	double*  eigVal = new double[ m_iDim ];
	double*  newV = new double[ m_iDim ];

	double** tempInvK = allocate2DDouble( m_iDim, m_iDim );
	for( i = 0; i < numPt; i++ )
	{
		inv( K[i], m_iDim, m_iDim, tempInvK );
		eig_sys( m_iDim, tempInvK, eigVec, eigVal );
		matConstMultiply( m_iDim, m_iDim, tempInvK, 1.0 / eigVal[0], tempInvK );
		copyMatrix( tempInvK, K[i], m_iDim, m_iDim );
	}
	safeDelete2DDouble( m_iDim, tempInvK );

	// neighborhood radius^2
	double radius2 = - m_fSigmaScale * log( m_fEpsBoundary );
	printf( "cut-off radius = %f\n", radius2 );

	printf( "create neighbor list .... ");
	// create neighbor list
	int** neighborList = new int*[numPt];
	int* neighborNum = new int[numPt];
	for( i = 0; i < numPt; i++ )
	{
		int numNeighbor = numPt;
		numNeighbor = annTree.annkFRSearch( (ANNpoint) x[i], (ANNdist) radius2, numNeighbor );
		int* tempIndexArray = new int[numNeighbor];
		numNeighbor = annTree.annkFRSearch( (ANNpoint) x[i], (ANNdist) radius2, numNeighbor, (ANNidxArray) tempIndexArray, (ANNdistArray) distArray );
		neighborList[i] = tempIndexArray;
		neighborNum[i] = numNeighbor;
	}
	printf( "Done\n" );

	printf( "Start EM Plane Fitting ... " );

	bool finished = false; 
	int iter = 0;
	while( !finished && (iter < m_iMaxIter) )
	{
		finished = true;

		//=========================================
		// E-Step
		//=========================================
		double distPlane2 = 0;
		double distvKv = 0;
		double prob = 0;
		for( i = 0; i < numPt; i++ )
		{
			distPlane2 = dotProduct( m_iDim, x[i], v );
			distPlane2 = distPlane2 * distPlane2;

			matVecMultiply( m_iDim, m_iDim, K[i], v, nVec );
			distvKv = fabs( dotProduct( m_iDim, v, nVec ) );

			prob = exp( - ( distPlane2 / (2.0 * sigma0) ) - ( distvKv / (2.0 * sigma1) ) );

			w[i] = prob;
		}

		//=========================================
		// M-Step
		//=========================================

		// Estimate alpha
		double newAlpha = 0;
		for( i = 0; i < numPt; i++ )
			newAlpha += w[i];
		newAlpha = newAlpha / double( numPt );
//		if( newAlpha < 0.01 ) newAlpha = 0.01;
//		alpha = newAlpha;

		// Estimate K[i] and sigma2
		int numNeighbor = 0;
		double newSigma2 = 0;
		double s2NormFactor = 0;
		for( i = 0; i < numPt; i++ )
		{
			memcpy( indexArray, neighborList[i], sizeof( int ) * neighborNum[i] ); 
			numNeighbor = neighborNum[i];

			zeroMat( m_iDim, m_iDim, newK );

			double normFactor = 0;
			for( j = 0; j < numNeighbor; j++ )
			{
				int index = indexArray[j];

				// find S here
				zeroMat( m_iDim, m_iDim, tempS );
				getSinv( x[i], x[index], K[index], tempS, m_fSigmaScale );

				newSigma2 += matFDist2( m_iDim, m_iDim, K[i], tempS ) * w[i] * w[index];

				matConstMultiply( m_iDim, m_iDim, tempS, w[index], tempS );
				addMat( m_iDim, m_iDim, newK, tempS, newK );

				normFactor += w[index];

				s2NormFactor += w[i] * w[index];
			}

			outerProduct( m_iDim, v, m_iDim, v, tempvv );
			matConstMultiply( m_iDim, m_iDim, tempvv, w[i] * sigma2 / (2.0 * sigma1), tempvv );
			subMat( m_iDim, m_iDim, newK, tempvv, newK );
			if( !nearzero( normFactor ) )
				matConstMultiply( m_iDim, m_iDim, newK, 1.0 / normFactor, newK );

			for( j = 0; j < m_iDim; j++ )
				newK[j][j] += 0.01;

			// normalize newK
			eig_sys( m_iDim, newK, eigVec, eigVal );
			matConstMultiply( m_iDim, m_iDim, newK, 1.0 / eigVal[0], newK );
			
			copyMatrix( newK, Kp[i], m_iDim, m_iDim );
		}
		if( !nearzero( s2NormFactor ) ) newSigma2 = newSigma2 / s2NormFactor;
		if( newSigma2 < m_afSigma[2] ) newSigma2 = m_afSigma[2];


		// Estimate v
		zeroMat( m_iDim, m_iDim, M );
		for( i = 0; i < numPt; i++ )
		{
			outerProduct( m_iDim, x[i], m_iDim, x[i], G );
			matConstMultiply( m_iDim, m_iDim, G, w[i], G );
			matConstMultiply( m_iDim, m_iDim, K[i], w[i] * sigma0 / sigma1, tempK );
			addMat( m_iDim, m_iDim, G, tempK, G );
			addMat( m_iDim, m_iDim, M, G, M ); 
		}
		
		eig_sys( m_iDim, M, eigVec, eigVal );
		memcpy( newV, eigVec[ m_iDim - 1 ], sizeof( double ) * m_iDim );
		

		// Estimate sigma0
		double newSigma0 = 0;
		double s0NormFactor = 0;
		for( i = 0; i < numPt; i++ )
		{
			double dist = dotProduct( m_iDim, x[i], v );
			newSigma0 += w[i] * dist * dist;
			s0NormFactor += w[i];
		}
		if( !nearzero( s0NormFactor ) ) newSigma0 = newSigma0 / s0NormFactor;
		if( newSigma0 < m_afSigma[0] ) newSigma0 = m_afSigma[0];



		// Estimate sigma1
		double newSigma1 = 0;
		double s1NormFactor = 0;
		for( i = 0; i < numPt; i++ )
		{
			memset( nVec, 0, sizeof( double ) * m_iDim );
			matVecMultiply( m_iDim, m_iDim, K[i], v, nVec );
			newSigma1 += w[i] * fabs( dotProduct( m_iDim, v, nVec ) );
			s1NormFactor += w[i];
		}
		if( !nearzero( s1NormFactor ) ) newSigma1 = newSigma1 / s1NormFactor;
		if( newSigma1 < m_afSigma[1] ) newSigma1 = m_afSigma[1];


		// check leaving condition
		for( i = 0; i < numPt; i++ )
			if( !nearzero( matFDist2( m_iDim, m_iDim, K[i], Kp[i] ) ) ) finished = false;

		if( !nearzero( alpha - newAlpha ) ) finished = false;
		if( !nearzero( vecDist2( m_iDim, v, newV ) ) ) finished = false;
		if( !nearzero( sigma0 - newSigma0 ) ) finished = false;
		if( !nearzero( sigma1 - newSigma1 ) ) finished = false;
		if( !nearzero( sigma2 - newSigma2 ) ) finished = false;

		// update parameters with newly estimated parameter
		for( i = 0; i < numPt; i++ )
			copyMatrix( Kp[i], K[i], m_iDim, m_iDim );
		alpha = newAlpha;
		memcpy( v, newV, sizeof( double ) * m_iDim );
		sigma0 = newSigma0;
		sigma1 = newSigma1;
		sigma2 = newSigma2;

		iter++;

	}

	printf( "Iteration = %d\n", iter );
	printf( "Final Sigmas:\n" );
	printf( "sigma0 = %f\n", sigma0 );
	printf( "sigma1 = %f\n", sigma1 );
	printf( "sigma2 = %f\n", sigma2 );

	SAFEDELARR( nVec );

	annClose();

	SAFEDELARR( indexArray );
	SAFEDELARR( distArray );
	safeDelete2DDouble( m_iDim, newK );
	safeDelete2DDouble( m_iDim, tempS );
	safeDelete2DDouble( m_iDim, tempvv );
	safeDelete2DDouble( m_iDim, M );
	safeDelete2DDouble( m_iDim, G );
	safeDelete2DDouble( m_iDim, tempK );
	safeDelete2DDouble( m_iDim, eigVec );
	SAFEDELARR( eigVal );
	SAFEDELARR( newV );

	if( Kp != NULL )
	{
		for( i = 0; i < numPt; i++ )
			safeDelete2DDouble( m_iDim, Kp[i] );
		SAFEDELARR( Kp );
	}

	if( neighborList != NULL )
		for( i = 0; i < numPt; i++ )
			SAFEDELARR( neighborList[i] );
	SAFEDELARR( neighborList );
	SAFEDELARR( neighborNum );


	return iter;
}


int EMPlane::em_plan_Fmatrix( double** x, int numPt, double*& v, double***& K, double*& w )
{

	assert( x != NULL );
	assert( v != NULL );
	assert( K != NULL );
	assert( w != NULL );

	int i, j, k;

	// allocate space for temporary storage
	double* nVec = new double[ m_iDim ];
	memset( nVec, 0, sizeof( double ) * m_iDim );

	double*** Kp = new double**[ numPt ];
	for( i = 0; i < numPt; i++ )
	{
		Kp[i] = allocate2DDouble( m_iDim, m_iDim );
		eye( m_iDim, Kp[i] );
	}

	// the paramter to be estimated in the algorithm
	double alpha = 0.9;
	double sigma0 = 100;
	double sigma1 = 100;
	double sigma2 = 100;

	// create ANN kd-tree
	ANNkd_tree annTree( (ANNpointArray) x, numPt, m_iDim );

	// create temporary storage
	int* indexArray = new int[ numPt ];
	double* distArray = new double[ numPt ];
	double** newK = allocate2DDouble( m_iDim, m_iDim );
	double** tempS = allocate2DDouble( m_iDim, m_iDim );
	double** tempvv = allocate2DDouble( m_iDim, m_iDim );
	double** M = allocate2DDouble( m_iDim, m_iDim );
	double** G = allocate2DDouble( m_iDim, m_iDim );
	double** tempK = allocate2DDouble( m_iDim, m_iDim );
	double** eigVec = allocate2DDouble( m_iDim, m_iDim );
	double*  eigVal = new double[ m_iDim ];
	double*  newV = new double[ m_iDim ];

	double** tempInvK = allocate2DDouble( m_iDim, m_iDim );
	for( i = 0; i < numPt; i++ )
	{
		inv( K[i], m_iDim, m_iDim, tempInvK );
		eig_sys( m_iDim, tempInvK, eigVec, eigVal );
		matConstMultiply( m_iDim, m_iDim, tempInvK, 1.0 / eigVal[0], tempInvK );
		copyMatrix( tempInvK, K[i], m_iDim, m_iDim );
	}
	safeDelete2DDouble( m_iDim, tempInvK );

	// neighborhood radius^2
	double radius2 = - m_fSigmaScale * log( m_fEpsBoundary );
	printf( "cut-off radius = %f\n", radius2 );

	// create neighbor list
	printf( "Create neighbor list .... " );
	int** neighborList = new int*[numPt];
	int* neighborNum = new int[numPt];
	for( i = 0; i < numPt; i++ )
	{
		int numNeighbor = numPt;
		numNeighbor = annTree.annkFRSearch( (ANNpoint) x[i], (ANNdist) radius2, numNeighbor );
		int* tempIndexArray = new int[numNeighbor];
		numNeighbor = annTree.annkFRSearch( (ANNpoint) x[i], (ANNdist) radius2, numNeighbor, (ANNidxArray) tempIndexArray, (ANNdistArray) distArray );
		neighborList[i] = tempIndexArray;
		neighborNum[i] = numNeighbor;
	}
	printf( "Done\n" );

	printf( "Start EM Plane Fitting ... " );

	bool firstIter = true;
	bool finished = false; 
	int iter = 0;
	while( !finished && (iter < m_iMaxIter) )
	{
		finished = true;

		//=========================================
		// E-Step
		//=========================================
		double distPlane2 = 0;
		double distvKv = 0;
		double prob = 0;
		
		if( !firstIter )
			for( i = 0; i < numPt; i++ )
			{
				distPlane2 = dotProduct( m_iDim, x[i], v );
				distPlane2 = distPlane2 * distPlane2;

				matVecMultiply( m_iDim, m_iDim, K[i], v, nVec );
				distvKv = fabs( dotProduct( m_iDim, v, nVec ) );

				prob = exp( - ( distPlane2 / (2.0 * sigma0) ) - ( distvKv / (2.0 * sigma1) ) );

				w[i] = prob;
			}
		else
		{
			zeroMat( m_iDim, m_iDim, M );
			for( i = 0; i < numPt; i++ )
			{
				outerProduct( m_iDim, x[i], m_iDim, x[i], G );
				matConstMultiply( m_iDim, m_iDim, G, w[i], G );
				addMat( m_iDim, m_iDim, M, G, M ); 
			}

			eig_sys( m_iDim, M, eigVec, eigVal );
			memcpy( v, eigVec[ m_iDim - 1 ], sizeof( double ) * m_iDim );
			
			FmatrixDiagSetZero( v );

			firstIter = false;
		}

		//=========================================
		// M-Step
		//=========================================

		// Estimate alpha
		double newAlpha = 0;
		for( i = 0; i < numPt; i++ )
			newAlpha += w[i];
		newAlpha = newAlpha / double( numPt );
		//		if( newAlpha < 0.01 ) newAlpha = 0.01;
		//		alpha = newAlpha;

		// Estimate K[i] and sigma2
		int numNeighbor = 0;
		double newSigma2 = 0;
		double s2NormFactor = 0;
		for( i = 0; i < numPt; i++ )
		{
			memcpy( indexArray, neighborList[i], sizeof( int ) * neighborNum[i] ); 
			numNeighbor = neighborNum[i];

			zeroMat( m_iDim, m_iDim, newK );

			double normFactor = 0;
			for( j = 0; j < numNeighbor; j++ )
			{
				int index = indexArray[j];

				// find S here
				zeroMat( m_iDim, m_iDim, tempS );
				getSinv( x[i], x[index], K[index], tempS, m_fSigmaScale );

				newSigma2 += matFDist2( m_iDim, m_iDim, K[i], tempS ) * w[i] * w[index];

				matConstMultiply( m_iDim, m_iDim, tempS, w[index], tempS );
				addMat( m_iDim, m_iDim, newK, tempS, newK );

				normFactor += w[index];
				s2NormFactor += w[i] * w[index];
			}

			outerProduct( m_iDim, v, m_iDim, v, tempvv );
			matConstMultiply( m_iDim, m_iDim, tempvv, w[i] * sigma2 / (2.0 * sigma1), tempvv );
			subMat( m_iDim, m_iDim, newK, tempvv, newK );
			if( !nearzero( normFactor ) )
				matConstMultiply( m_iDim, m_iDim, newK, 1.0 / normFactor, newK );

			for( j = 0; j < m_iDim; j++ )
				newK[j][j] += 0.01;

			// normalize newK
			eig_sys( m_iDim, newK, eigVec, eigVal );
			matConstMultiply( m_iDim, m_iDim, newK, 1.0 / eigVal[0], newK );

			copyMatrix( newK, Kp[i], m_iDim, m_iDim );
		}
		if( !nearzero( s2NormFactor ) ) newSigma2 = newSigma2 / s2NormFactor;
		if( newSigma2 < m_afSigma[2] ) newSigma2 = m_afSigma[2];


		// Estimate v
		zeroMat( m_iDim, m_iDim, M );
		for( i = 0; i < numPt; i++ )
		{
			outerProduct( m_iDim, x[i], m_iDim, x[i], G );
			matConstMultiply( m_iDim, m_iDim, G, w[i], G );
			matConstMultiply( m_iDim, m_iDim, K[i], w[i] * w[i] * sigma0 / sigma1, tempK );
			addMat( m_iDim, m_iDim, G, tempK, G );
			addMat( m_iDim, m_iDim, M, G, M ); 
		}

		eig_sys( m_iDim, M, eigVec, eigVal );
		memcpy( newV, eigVec[ m_iDim - 1 ], sizeof( double ) * m_iDim );
		
		FmatrixDiagSetZero( newV );


		// Estimate sigma0
		double newSigma0 = 0;
		double s0NormFactor = 0;
		for( i = 0; i < numPt; i++ )
		{
			double dist = dotProduct( m_iDim, x[i], v );
			newSigma0 += w[i] * dist * dist;
			s0NormFactor += w[i];
		}
		if( !nearzero( s0NormFactor ) ) newSigma0 = newSigma0 / s0NormFactor;
		if( newSigma0 < m_afSigma[0] ) newSigma0 = m_afSigma[0];



		// Estimate sigma1
		double newSigma1 = 0;
		double s1NormFactor = 0;
		for( i = 0; i < numPt; i++ )
		{
			memset( nVec, 0, sizeof( double ) * m_iDim );
			matVecMultiply( m_iDim, m_iDim, K[i], v, nVec );
			newSigma1 += w[i] * fabs( dotProduct( m_iDim, v, nVec ) );
			s1NormFactor += w[i];
		}
		if( !nearzero( s1NormFactor ) ) newSigma1 = newSigma1 / s1NormFactor;
		if( newSigma1 < m_afSigma[1] ) newSigma1 = m_afSigma[1];


		// check leaving condition
		for( i = 0; i < numPt; i++ )
			if( !nearzero( matFDist2( m_iDim, m_iDim, K[i], Kp[i] ) ) ) finished = false;

		if( !nearzero( alpha - newAlpha ) ) finished = false;
		if( !nearzero( vecDist2( m_iDim, v, newV ) ) ) finished = false;
		if( !nearzero( sigma0 - newSigma0 ) ) finished = false;
		if( !nearzero( sigma1 - newSigma1 ) ) finished = false;
		if( !nearzero( sigma2 - newSigma2 ) ) finished = false;

		// update parameters with newly estimated parameter
		for( i = 0; i < numPt; i++ )
			copyMatrix( Kp[i], K[i], m_iDim, m_iDim );
		alpha = newAlpha;
		memcpy( v, newV, sizeof( double ) * m_iDim );
		sigma0 = newSigma0;
		sigma1 = newSigma1;
		sigma2 = newSigma2;

		iter++;

	}

	printf( "Iteration = %d\n", iter );

	printf( "Final Sigmas:\n" );
	printf( "sigma0 = %f\n", sigma0 );
	printf( "sigma1 = %f\n", sigma1 );
	printf( "sigma2 = %f\n", sigma2 );


	SAFEDELARR( nVec );

//	annClose();

	SAFEDELARR( indexArray );
	SAFEDELARR( distArray );
	safeDelete2DDouble( m_iDim, newK );
	safeDelete2DDouble( m_iDim, tempS );
	safeDelete2DDouble( m_iDim, tempvv );
	safeDelete2DDouble( m_iDim, M );
	safeDelete2DDouble( m_iDim, G );
	safeDelete2DDouble( m_iDim, tempK );
	safeDelete2DDouble( m_iDim, eigVec );
	SAFEDELARR( eigVal );
	SAFEDELARR( newV );

	if( Kp != NULL )
	{
		for( i = 0; i < numPt; i++ )
			safeDelete2DDouble( m_iDim, Kp[i] );
		SAFEDELARR( Kp );
	}

	if( neighborList != NULL )
		for( i = 0; i < numPt; i++ )
			SAFEDELARR( neighborList[i] );
	SAFEDELARR( neighborList );
	SAFEDELARR( neighborNum );


	return iter;
}


void EMPlane::initialize( int numPt, double* v, double** x, double*** K, int iter )
{
	// create ANN kd-tree
	ANNkd_tree annTree( (ANNpointArray) x, numPt, m_iDim );

	int i, j, k;

	int* indexArray = new int[ numPt ];
	double* distArray = new double[ numPt ];
	double** newK = allocate2DDouble( m_iDim, m_iDim );
	double** tempS = allocate2DDouble( m_iDim, m_iDim );
	double** M = allocate2DDouble( m_iDim, m_iDim );
	double** eigVec = allocate2DDouble( m_iDim, m_iDim );
	double*  eigVal = new double[ m_iDim ];
	double*  newV = new double[ m_iDim ];

	double*** Kp = new double**[ numPt ];
	for( i = 0; i < numPt; i++ )
		Kp[i] = allocate2DDouble( m_iDim, m_iDim );


	double radius2 = - m_fSigmaScale * log( m_fEpsBoundary );

	for( k = 0; k < iter; k++ )
	{
		// Estimate K[i]
		int numNeighbor = 0;
		for( i = 0; i < numPt; i++ )
		{

			numNeighbor = numPt;
			numNeighbor = annTree.annkFRSearch( (ANNpoint) x[i], (ANNdist) radius2, numNeighbor );
			numNeighbor = annTree.annkFRSearch( (ANNpoint) x[i], (ANNdist) radius2, numNeighbor, (ANNidxArray) indexArray, (ANNdistArray) distArray );
			
			zeroMat( m_iDim, m_iDim, newK );

			for( j = 0; j < numNeighbor; j++ )
			{
				int index = indexArray[j];

				// find S here
				zeroMat( m_iDim, m_iDim, tempS );

				getS( x[i], x[index], K[index], tempS, m_fSigmaScale );

				addMat( m_iDim, m_iDim, newK, tempS, newK );

			}

			for( j = 0; j < m_iDim; j++ )
				newK[j][j] += 0.01;
			
			copyMatrix( newK, Kp[i], m_iDim, m_iDim );
		}

		for( i = 0; i < numPt; i++ )
			copyMatrix( Kp[i], K[i], m_iDim, m_iDim );
	}

	// estimate v
	zeroMat( m_iDim, m_iDim, M );
	for( i = 0; i < numPt; i++ )
		addMat( m_iDim, m_iDim, M, K[i], M ); 
		
	eig_sys( m_iDim, M, eigVec, eigVal );
	memcpy( v, eigVec[ 0 ], sizeof( double ) * m_iDim );

	// normalize K
	for( i = 0; i < numPt; i++ ) 
	{
		eig_sys( m_iDim, K[i], eigVec, eigVal );
		matConstMultiply( m_iDim, m_iDim, K[i], 1.0 / eigVal[0], K[i] );
	}
	
	annClose();

	SAFEDELARR( indexArray );
	SAFEDELARR( distArray );
	safeDelete2DDouble( m_iDim, newK );
	safeDelete2DDouble( m_iDim, tempS );
	safeDelete2DDouble( m_iDim, M );
	safeDelete2DDouble( m_iDim, eigVec );
	SAFEDELARR( eigVal );

	if( Kp != NULL )
	{
		for( i = 0; i < numPt; i++ )
			safeDelete2DDouble( m_iDim, Kp[i] );
		SAFEDELARR( Kp );
	}
}


void EMPlane::getS( double* x1, double* x2, double** K2, double** S, double sigma )
{
	assert( x1 != NULL );
	assert( x2 != NULL );
	assert( K2 != NULL );
	assert( S  != NULL );

	int i, j;

	if( R == NULL ) R = allocate2DDouble( m_iDim, m_iDim );
	if( Rp == NULL ) Rp = allocate2DDouble( m_iDim, m_iDim );
	if( outr == NULL ) outr = allocate2DDouble( m_iDim, m_iDim );
	if( r == NULL ) r = new double[ m_iDim ];
	if( I == NULL ) I = allocate2DDouble( m_iDim, m_iDim );
	eye( m_iDim, I );

	for( i = 0; i < m_iDim; i++ )
		r[i] = x1[i] - x2[i];

	double length_r2 = norm2( m_iDim, r );
	double length_r  = sqrt( length_r2 );

	if( !nearzero( length_r ) )
		for( i = 0; i < m_iDim; i++ )
			r[i] = r[i] / length_r;

	// outer product of r
	outerProduct( m_iDim, r, m_iDim, r, outr ); 

	for( i = 0; i < m_iDim; i++ )
		for( j = 0; j < m_iDim; j++ )
			outr[i][j] = 2.0 * outr[i][j];

	// find R
	subMat( m_iDim, m_iDim, I, outr, R );

	for( i = 0; i < m_iDim; i++ )
		for( j = 0; j < m_iDim; j++ )
			outr[i][j] = outr[i][j] / 4.0;

	// find Rp
	subMat( m_iDim, m_iDim, I, outr, Rp );
	multiMat( m_iDim, m_iDim, m_iDim, Rp, R, outr );
	copyMatrix( outr, Rp, m_iDim, m_iDim );

	//find S
	multiMat( m_iDim, m_iDim, m_iDim, R, K2, outr );
	multiMat( m_iDim, m_iDim, m_iDim, outr, Rp, S );
	
	double c = exp( - fabs( length_r2 ) / sigma );

	// scale the S by distance
	for( i = 0; i < m_iDim; i++ )
		for( j = 0; j < m_iDim; j++ )
			S[i][j] = c * S[i][j];
}

void EMPlane::getSinv( double* x1, double * x2, double** K2inv, double** Sinv, double sigma )
{
	assert( x1 != NULL );
	assert( x2 != NULL );
	assert( K2inv != NULL );
	assert( Sinv  != NULL );

	int i, j;

	if( R == NULL ) R = allocate2DDouble( m_iDim, m_iDim );
	if( Rp == NULL ) Rp = allocate2DDouble( m_iDim, m_iDim );
	if( outr == NULL ) outr = allocate2DDouble( m_iDim, m_iDim );
	if( r == NULL ) r = new double[ m_iDim ];
	if( I == NULL ) I = allocate2DDouble( m_iDim, m_iDim );
	eye( m_iDim, I );

	for( i = 0; i < m_iDim; i++ )
		r[i] = x1[i] - x2[i];

	double length_r2 = norm2( m_iDim, r );
	double length_r  = sqrt( length_r2 );

	if( !nearzero( length_r ) )
		for( i = 0; i < m_iDim; i++ )
			r[i] = r[i] / length_r;

	// outer product of r
	outerProduct( m_iDim, r, m_iDim, r, outr ); 

	for( i = 0; i < m_iDim; i++ )
		for( j = 0; j < m_iDim; j++ )
			outr[i][j] = 2.0 * outr[i][j];

	// find R
	subMat( m_iDim, m_iDim, I, outr, R );

	for( i = 0; i < m_iDim; i++ )
		for( j = 0; j < m_iDim; j++ )
			outr[i][j] =  outr[i][j] / 2.0;

	// find Rp
	addMat( m_iDim, m_iDim, I, outr, Rp );
	multiMat( m_iDim, m_iDim, m_iDim, R, Rp, outr );
	copyMatrix( outr, Rp, m_iDim, m_iDim );

	//find S
	multiMat( m_iDim, m_iDim, m_iDim, Rp, K2inv, outr );
	multiMat( m_iDim, m_iDim, m_iDim, outr, R, Sinv );

	double c = 1.0 / exp( - fabs( length_r2 ) / sigma );

	// scale the Sinv by distance
	for( i = 0; i < m_iDim; i++ )
		for( j = 0; j < m_iDim; j++ )
			Sinv[i][j] = Sinv[i][j] * c;
}

int FmatrixDiagSetZero( double* f )
{
	assert( f != NULL );
	
	int i, j, k;
	
	double** a = allocate2DDouble( 3, 3 );
	double*  w = new double[3];
	double** v = allocate2DDouble( 3, 3 );
	double** result = allocate2DDouble( 3, 3 );
	
	a[0][0] = f[0];
	a[1][0] = f[1];
	a[2][0] = f[2];
	
	a[0][1] = f[3];
	a[1][1] = f[4];
	a[2][1] = f[5];
	
	a[0][2] = f[6];
	a[1][2] = f[7];
	a[2][2] = f[8];

	svdcmp( a, 3, 3, w, v );
	
	int small_index = 0;
	for( i = 1; i < 3; i++ )
		if( w[i] < w[small_index] ) small_index = i;

	w[small_index] = 0;
	
	for( i = 0; i < 3; i++ )
		for( j = 0; j < 3; j++ )
			a[i][j] = a[i][j] * w[j];
			
	for( i = 0; i < 3; i++ )
		memset( result[i], 0, sizeof( double ) * 3 );
			
	for( i = 0; i < 3; i++ )
		for( j = 0; j < 3; j++ )
			for( k = 0; k < 3; k++ ) 
				result[i][j] += a[i][k] * v[j][k];
			
	
	f[0] = result[0][0];
	f[1] = result[1][0];
	f[2] = result[2][0];
	
	f[3] = result[0][1];
	f[4] = result[1][1];
	f[5] = result[2][1];
	
	f[6] = result[0][2];
	f[7] = result[1][2];
	f[8] = result[2][2];
	
	
	safeDelete2DDouble( 3, a );
	SAFEDELARR( w );
	safeDelete2DDouble( 3, v );
	safeDelete2DDouble( 3, result );

	return 1;
}

double** readFromFile( const char* fileName, int& numPt, int& dim )
{	
	FILE* fid = fopen( fileName, "r" );

	if( fid == NULL ) return NULL;
	
	fscanf( fid, "%d %d", &numPt, &dim );
	double** x = allocate2DDouble( numPt, dim ); 

	int i, j;
	for( i = 0; i < numPt; i++ )
		for( j = 0; j < dim; j++ )
			fscanf( fid, "%lf", &x[i][j] );

	fclose( fid );

	return x;
}

